import logging
import pytest
from cfn_docgen.adapters.cfn_specification_repository import CfnSpecificationRepository
from cfn_docgen.adapters.internal.cache import LocalFileCache
from cfn_docgen.adapters.internal.file_loader import RemoteFileLoader
from cfn_docgen.config import AppConfig, AppContext
from cfn_docgen.domain.model.cfn_specification import CfnSpecificationPropertyTypeName, CfnSpecificationResourceTypeName


@pytest.fixture
def cfn_specification_url():
    return "https://d1uauaxba7bl26.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json"

@pytest.fixture
def context():
    return AppContext(log_level=logging.DEBUG)

@pytest.fixture
def repository(context:AppContext):
    return CfnSpecificationRepository(
        source_url="https://d1uauaxba7bl26.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        loader=RemoteFileLoader(context=context),
        cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=context),
        recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
        context=context
    )


def test_CfnSpecificationRepository_get_resource_spec(
    repository:CfnSpecificationRepository,
    context:AppContext,
):

    for resource_type in repository.spec.ResourceTypes.keys():
        resource_spec = repository.get_resource_spec(
            CfnSpecificationResourceTypeName(resource_type, context),
        )
        assert resource_spec is not None
        if resource_spec.Documentation is not None:
            assert (
                resource_spec.Documentation.startswith("http://") or resource_spec.Documentation.startswith("https://")
            ) , f"{resource_type}/{resource_spec.Documentation}"
        
        for prop_name, prop in resource_spec.Properties.items():
            if prop.Documentation is not None:
                assert (
                    prop.Documentation.startswith("http://") or prop.Documentation.startswith("https://")
                ), f"{resource_type}/{prop_name}/{prop.Documentation}/"

            if prop.PrimitiveItemType is not None or prop.ItemType is not None:
                assert prop.Type is not None and (prop.Type == "List" or prop.Type == "Map"), f"{resource_type}/{prop_name}/{prop}"
            if prop.Type is not None and (prop.Type != "List" and prop.Type != "Map"):
                assert (
                    prop.PrimitiveItemType is None and prop.ItemType is None and prop.PrimitiveType is None
                ), f"{resource_type}/{prop_name}/{prop}"

def test_CfnSpecificationRepository_get_resource_spec_KeyError(
    repository:CfnSpecificationRepository,
    context:AppContext,
):
    invalid_key = CfnSpecificationResourceTypeName("AWS::Invalid::Key", context)
    with pytest.raises(KeyError) as ex:
        repository.get_resource_spec(invalid_key)
    assert invalid_key.fullname in ex.value.args 

def test_CfnSpecificationRepository_get_property_spec_KeyError(
    repository:CfnSpecificationRepository,
    context:AppContext,
):
    invalid_key = CfnSpecificationPropertyTypeName("AWS::Invalid::Key.1", context)
    with pytest.raises(KeyError) as ex:
        repository.get_property_spec(invalid_key)
    assert invalid_key.fullname in ex.value.args 


def test_CfnSpecificationRepository_get_property_spec(
    repository:CfnSpecificationRepository,
    context:AppContext,
):
    for property_type in repository.spec.PropertyTypes.keys():
        property_spec = repository.get_property_spec(CfnSpecificationPropertyTypeName(property_type, context))
        assert property_spec is not None
        if property_spec.Documentation is not None:
            assert (
                property_spec.Documentation.startswith("http://") or property_spec.Documentation.startswith("https://")
            ), f"{property_type}/{property_spec.Documentation}"

        if property_spec.Properties is None:
            assert (
                property_spec.ItemType is not None or
                property_spec.PrimitiveItemType is not None or
                property_spec.PrimitiveType is not None or
                property_spec.Type is not None
            ) , f"{property_type}/{property_spec}"
        
        else:
            for prop_name, prop in property_spec.Properties.items():
                if prop.Documentation is not None:
                    assert (
                        prop.Documentation.startswith("http://") or prop.Documentation.startswith("https://")
                    ), f"{property_type}/{prop_name}/{prop.Documentation}/"

                if prop.PrimitiveItemType is not None or prop.ItemType is not None:
                    assert prop.Type is not None and (prop.Type == "List" or prop.Type == "Map"), f"{property_type}/{prop_name}/{prop}"
                if prop.Type is not None and (prop.Type != "List" and prop.Type != "Map"):
                    assert (
                        prop.PrimitiveItemType is None and prop.ItemType is None and prop.PrimitiveType is None
                    ), f"{property_type}/{prop_name}/{prop}"

def test_CfnSpecificationRepository_list_properties_for_resource(
    repository:CfnSpecificationRepository,
    context:AppContext,
):
    resource_type = CfnSpecificationResourceTypeName("AWS::EC2::Instance", context)
    proeprties = repository.list_properties_for_resource(resource_type)
    assert proeprties.get(
        CfnSpecificationPropertyTypeName(f"{resource_type.fullname}.BlockDeviceMapping", context).fullname
    )
    assert proeprties.get(
        CfnSpecificationPropertyTypeName(f"{resource_type.fullname}.Ebs", context).fullname
    )

def test_CfnSpecificationRepository_get_specs_for_resources(
    repository:CfnSpecificationRepository,
    context:AppContext,
):
    resource_type = CfnSpecificationResourceTypeName("AWS::EC2::Instance", context)
    specs = repository.get_specs_for_resource(resource_type)
    assert specs.PropertySpecs.get(
        CfnSpecificationPropertyTypeName(f"{resource_type.fullname}.BlockDeviceMapping", context).fullname
    )
    assert specs.PropertySpecs.get(
        CfnSpecificationPropertyTypeName(f"{resource_type.fullname}.Ebs", context).fullname
    )

@pytest.mark.parametrize("resource_type,expected", [
    ("AWS::WAFv2::RuleGroup", True), ("AWS::EC2::Instance", False)
])
def test_CfnSpecificationRepository_is_recursive(
    resource_type:str, expected:bool, repository:CfnSpecificationRepository,
    context:AppContext,
):
    assert repository.is_recursive(
        CfnSpecificationResourceTypeName(resource_type, context)
    ) == expected



import logging
import sys
import click
from cfn_docgen.adapters.cfn_document_storage import document_storage_facotory
from cfn_docgen.adapters.cfn_specification_repository import CfnSpecificationRepository
from cfn_docgen.adapters.cfn_template_provider import template_provider_factory
from cfn_docgen.adapters.internal.cache import LocalFileCache
from cfn_docgen.adapters.internal.file_loader import RemoteFileLoader, file_loader_factory
from cfn_docgen.config import AppConfig, AppContext
from cfn_docgen.domain.model.cfn_document_generator import document_generator_factory
from cfn_docgen.domain.services.cfn_docgen_service import CfnDocgenService

from cfn_docgen.entrypoints.cli.model.cli_model import CfnDocgenCLIUnitsOfWork, CliArguement, SupportedFormat

@click.group(name="cfn-docgen")
@click.version_option(package_name="cfn-docgen")
def main():
    pass


@main.command()
@click.option(
    "-f", "--format", "fmt", required=True, type=click.Choice(["markdown"]), 
    help="format of output file"
)
@click.option(
    "-s", "--source", "source", required=True, type=str, 
    help="input cfn template source. (can be file or directory with form of local path or S3 url)"
)
@click.option(
    "-d", "--dest", "dest", required=True, type=str, 
    help="output document destination. (can be file or directory with form of local path or S3 url)"
)
@click.option(
    "--debug", "debug", required=False, is_flag=True, show_default=True, default=False,
    help="enable logging"
)
def docgen(fmt:SupportedFormat, source:str, dest:str, debug:bool):
    context = AppContext(
        request_id=None,
        logger_name="cfn-docgen",
        log_level=logging.DEBUG if debug else logging.CRITICAL,
    )
    args = CliArguement(
        subcommand="docgen",
        format=fmt,
        source=source,
        dest=dest,
    )
    context.log_debug(f"received args [{' '.join(args.as_list())}]")

    try:
        units_of_work = CfnDocgenCLIUnitsOfWork(
            args=args,
            file_loader_factory=file_loader_factory,
            context=context,
        )
    except AssertionError:
        context.log_error("failed to setup pairs of template sources and document dests")
        click.echo(context.log_messages.as_string(logging.INFO))
        sys.exit(1)

    try:
        service = CfnDocgenService(
            cfn_template_provider_facotry=template_provider_factory,
            cfn_document_generator_factory=document_generator_factory,
            cfn_document_storage_factory=document_storage_facotory,
            cfn_specification_repository=CfnSpecificationRepository(
                source_url=AppConfig.DEFAULT_SPECIFICATION_URL,
                loader=RemoteFileLoader(context=context),
                cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=context),
                recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
                context=context,
            ),
            context=context,
        )
        
    except Exception:
        context.log_error("failed to setup CfnDocgenService")
        click.echo(context.log_messages.as_string(logging.INFO))
        sys.exit(1)

    for command_input in units_of_work.provide():
        try:
            context.log_debug(f"start cfn-docgen process for template source [{command_input.template_source.source}] and document dest [{command_input.document_dest.dest}]")
            result = service.main(command_input=command_input)
            context.log_info(f"successfully generate document [{result.document_dest}] from template [{command_input.template_source.source}]")
        except Exception:
            context.log_warning(f"failed to generate document [{command_input.document_dest.dest}] from template [{command_input.template_source.source}]")
            continue


    click.echo(context.log_messages.as_string(logging.INFO))
    sys.exit(0)



import os
from typing import List

from pydantic import BaseModel
from dataclasses import dataclass
from cfn_docgen.config import AppContext

from cfn_docgen.domain.model.cfn_document_generator import CfnDocumentDestination
from cfn_docgen.domain.model.cfn_template import CfnTemplateSource
from cfn_docgen.domain.services.cfn_docgen_service import CfnDocgenServiceCommandInput, SupportedFormat, ext_by_format

class S3NotificationEventRecordS3Object(BaseModel):
    key: str
class S3NotificationEventRecordS3Bucket(BaseModel):
    name: str
class S3NotificationEventRecordS3(BaseModel):
    bucket: S3NotificationEventRecordS3Bucket
    object: S3NotificationEventRecordS3Object
class S3NotificationEventRecord(BaseModel):
    s3: S3NotificationEventRecordS3
class S3NotificationEvent(BaseModel):
    Records: List[S3NotificationEventRecord]



@dataclass
class ServerlessArguement:
    format: SupportedFormat
    sources:List[S3NotificationEventRecordS3]
    dest_bucket:str
    dest_prefix:str
    
class CfnDocgenServerlessUnitsOfWork:
    def __init__(
        self,
        args:ServerlessArguement,
        context:AppContext,
    ) -> None:
        self.units_of_work = self.build_units_of_work(
            args=args,
            context=context,
        )
        try:
            assert len(self.units_of_work) > 0, "no valid template sources and document dests are proided"
        except AssertionError as ex:
            context.log_error(ex.args[0])
            raise ex

    def build_units_of_work(
        self, 
        args:ServerlessArguement, 
        context:AppContext,
    ) -> List[CfnDocgenServiceCommandInput]:
        
        units_of_work:List[CfnDocgenServiceCommandInput] = []

        try:
            # source must be a single file
            for source in args.sources:
                bucket = source.bucket.name
                key = source.object.key

                source_dir = os.path.dirname(key)
                base_prefix, _ = os.path.splitext(os.path.basename(key))
                dest_bucket = args.dest_bucket
                dest_prefix = args.dest_prefix
                if dest_prefix.endswith("/"):
                    dest_prefix = dest_prefix[:-1]

                units_of_work.append(
                    CfnDocgenServiceCommandInput(
                        template_source=CfnTemplateSource(f"s3://{bucket}/{key}", context=context),
                        document_dest=CfnDocumentDestination(
                            f"s3://{dest_bucket}/{dest_prefix}/{source_dir}/{base_prefix}.{ext_by_format[args.format]}",
                            context=context,
                        ),
                        fmt=args.format,
                    )
                )
            return units_of_work
        except Exception:
            context.log_error("failed to prepare template sources and document dests")
            return []

    def provide(self) -> List[CfnDocgenServiceCommandInput]:
        return self.units_of_work



from __future__ import annotations
from dataclasses import dataclass
import datetime
from logging import Logger
import logging
import os
from typing import Iterable, List, Optional
import uuid

class AppConfig:
    APP_NAME="cfn-docgen"
    APP_ROOT_DIR=os.path.join(
        os.path.expanduser("~"), f".{APP_NAME}"
    )
    CACHE_ROOT_DIR=os.path.join(APP_ROOT_DIR, "cache")
    DEFAULT_SPECIFICATION_URL="https://d1uauaxba7bl26.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json"
    RECURSIVE_RESOURCE_TYPES=[
        "AWS::WAFv2::WebACL",
        "AWS::WAFv2::RuleGroup",
        "AWS::AmplifyUIBuilder::Component",
        "AWS::IoTTwinMaker::ComponentType",
    ]

@dataclass
class AppContextLogMessage:
    level: int
    message: str
    timestamp: float
@dataclass
class AppContextLogMessages:
    messages: List[AppContextLogMessage]

    def as_string(self, level:int, delimeter:str="\n") -> str:
        prefixes = {
            logging.INFO: "[INFO]",
            logging.WARNING: "[WARNING]",
            logging.ERROR: "[ERROR]",
            logging.DEBUG: "[DEBUG]",
        }
        return_messages:Iterable[str] = map(
            lambda m: f"{prefixes[m.level]} {m.message}",
            filter(
                lambda m: m.level >= level, 
                sorted(self.messages, key=lambda m: m.timestamp),
            )
        )
        return delimeter.join(return_messages)

class AppContext:

    def __init__(
        self, 
        request_id:Optional[str]=None,
        logger_name:str="cfn-docgen", 
        log_level:int=logging.CRITICAL, 
    ) -> None:        
        self.request_id = str(uuid.uuid4()) if request_id is None else request_id
        self.__logger = AppLogger(name=logger_name, loglevel=log_level, stacklevel=3)
        self.log_messages = AppContextLogMessages(messages=[])

    def log_info(self, msg:str, ):
        now = datetime.datetime.now().timestamp()
        self.log_messages.messages.append(AppContextLogMessage(
            level=logging.INFO, message=msg, timestamp=now,
        ))
        self.__logger.info(msg)

    def log_warning(self, msg:str):
        now = datetime.datetime.now().timestamp()
        self.log_messages.messages.append(AppContextLogMessage(
            level=logging.WARNING, message=msg, timestamp=now,
        ))
        self.__logger.warning(msg)

    def log_error(self, msg:str):
        now = datetime.datetime.now().timestamp()
        self.log_messages.messages.append(AppContextLogMessage(
            level=logging.ERROR, message=msg, timestamp=now,
        ))
        self.__logger.error(msg)

    def log_debug(self, msg:str):
        now = datetime.datetime.now().timestamp()
        self.log_messages.messages.append(AppContextLogMessage(
            level=logging.DEBUG, message=msg, timestamp=now,
        ))
        self.__logger.debug(msg)

class AppLogger:
    __logger: Logger
    def __init__(self, name:str, loglevel:int, stacklevel:int) -> None:
        logger = logging.getLogger(name)
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(module)s - %(funcName)s - %(lineno)d - %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(loglevel)

        self.__logger = logger
        self.__stacklevel = stacklevel

    def info(self, msg:str, ):
        self.__logger.info(msg, stacklevel=self.__stacklevel)
    def warning(self, msg:str):
        self.__logger.warning(msg, stacklevel=self.__stacklevel, exc_info=True)
    def error(self, msg:str):
        self.__logger.error(msg, stacklevel=self.__stacklevel, exc_info=True)
    def debug(self, msg:str):
        self.__logger.debug(msg, stacklevel=self.__stacklevel)


import glob
import os
from typing import List, Optional
from urllib.parse import urlparse
import requests
import boto3
from cfn_docgen.config import AppContext
from cfn_docgen.domain.model.cfn_document_generator import CfnDocumentDestination
from cfn_docgen.domain.model.cfn_template import CfnTemplateSource # type: ignore
from cfn_docgen.domain.ports.internal.file_loader import IFileLoader

def document_loader_factory(
    document_dest:CfnDocumentDestination,
    context:AppContext,
) -> IFileLoader:
    match document_dest.type:
        case "LocalFilePath":
            context.log_debug(f"document dest type is [{document_dest.type}]. return LocalFileLoader")
            return LocalFileLoader(context)
        case "S3BucketKey":
            context.log_debug(f"document dest type is [{document_dest.type}]. return S3FileLoader")
            return S3FileLoader(context)
        case "HttpUrl":
            context.log_debug(f"document dest type is [{document_dest.type}]. return RemoteFileLoader")
            return RemoteFileLoader(context)
        case _: # type: ignore
            raise NotImplementedError(f"document dest type [{document_dest.type}] is not supported")


def file_loader_factory(
    template_source:CfnTemplateSource,
    context:AppContext,
) -> IFileLoader:
    match template_source.type:
        case "LocalFilePath":
            context.log_debug(f"template source type is [{template_source.type}]. return LocalFileLoader")
            return LocalFileLoader(context)
        case "S3BucketKey":
            context.log_debug(f"template source type is [{template_source.type}]. return S3FileLoader")
            return S3FileLoader(context)
        case "HttpUrl":
            context.log_debug(f"template source type is [{template_source.type}]. return RemoteFileLoader")
            return RemoteFileLoader(context)
        case _: # type: ignore
            raise NotImplementedError(f"template source type [{template_source.type}] is not supported")


class LocalFileLoader(IFileLoader):
    def __init__(self, context: AppContext) -> None:
        super().__init__(context)

    def download(self, source: str) -> bytes:
        with open(source, "rb") as fp:
            raw = fp.read()
        self.context.log_debug(f"download from [{source}]")
        return raw
    
    def upload(self, body: bytes, dest: str) -> None:
        # meke suer directory is exist
        os.makedirs(os.path.dirname(dest), exist_ok=True)
        with open(dest, "wb") as fp:
            fp.write(body)
        self.context.log_debug(f"upload to [{dest}]")

    def list(self, source: str) -> List[str]:
        if os.path.isfile(source):
            self.context.log_debug(f"[{source}] is a single file")
            return [source]
        else:
            files_and_dir = glob.glob(os.path.join(source, "**"), recursive=True)
            self.context.log_debug(f"[{source}] is a directory. listed files are [{'.'.join(files_and_dir)}]")
            return sorted([f for f in files_and_dir if os.path.isfile(f)])

class RemoteFileLoader(IFileLoader):

    def __init__(self, context: AppContext) -> None:
        super().__init__(context)

    def download(self, source: str) -> bytes:
        res = requests.get(source, timeout=10)
        self.context.log_debug(f"download from [{source}]")
        return res.content

    def upload(self, body: bytes, dest: str) -> None:
        raise NotImplementedError
    
    def list(self, source: str) -> List[str]:
        raise NotImplementedError
    
class S3FileLoader(IFileLoader):
    def __init__(self, context: AppContext) -> None:
        super().__init__(context)
        self.client = boto3.client("s3") # type: ignore

    def download(self, source: str) -> bytes:
        s3_url = urlparse(source)
        bucket = s3_url.netloc
        key = s3_url.path
        if key.startswith("/"):
            key = key[1:]
        res = self.client.get_object(
            Bucket=bucket,
            Key=key,
        )
        self.context.log_debug(f"download from [{source}]")
        return res["Body"].read()
    
    def upload(self, body: bytes, dest: str) -> None:
        s3_url = urlparse(dest)
        bucket = s3_url.netloc
        key = s3_url.path
        if key.startswith("/"):
            key = key[1:]
        self.client.put_object(
            Bucket=bucket,
            Key=key,
            Body=body,
        )
        self.context.log_debug(f"upload to [{dest}]")

    def list(self, source: str) -> List[str]:
        sources:List[str] = []

        source_url = urlparse(source)
        bucket = source_url.netloc
        prefix_or_key = source_url.path
        if prefix_or_key.startswith("/"):
            prefix_or_key = prefix_or_key[1:]

        # if source is a object key, return it
        try:
            res = self.client.head_object(
                Bucket=bucket,
                Key=prefix_or_key
            )
            self.context.log_debug(f"[{source}] is a single object")
            return [source]
        except Exception as ex:
            if "404" in str(ex):
                pass
            else:
                raise ex

        # if source is a prefix, return all keys
        prefix = prefix_or_key
        if not prefix.endswith("/"):
            prefix += "/"
        res = self.client.list_objects_v2(
            Bucket=bucket, Prefix=prefix
        )
        contents = res.get("Contents", None)
        if contents is None: # type: ignore
            return sources
        for content in contents:
            sources.append(
                os.path.join(f"s3://{bucket}",  content["Key"]) # type: ignore
            )
        next_token:Optional[str] = res.get("NextContinuationToken", None)
        while next_token is not None: # type: ignore
            res = self.client.list_objects_v2(
                Bucket=bucket, Prefix=prefix,
                ContinuationToken=next_token,
            )
            contents = res.get("Contents", None)
            if contents is None: # type: ignore
                return sources
            for content in contents:
                sources.append(
                    os.path.join(f"s3://{bucket}", content["Key"]) # type: ignore
                )
            next_token:str|None = res.get("NextContinuationToken", None)

        self.context.log_debug(f"[{source}] is a directory. listed keys are [{'.'.join(sources)}]")
        return sources


# pylint: disable=W0613
import logging
import os
from typing import Any, List, Mapping, Optional
from cfn_docgen.adapters.cfn_document_storage import document_storage_facotory
from cfn_docgen.adapters.cfn_specification_repository import CfnSpecificationRepository
from cfn_docgen.adapters.cfn_template_provider import template_provider_factory
from cfn_docgen.adapters.internal.cache import LocalFileCache
from cfn_docgen.adapters.internal.file_loader import RemoteFileLoader
from cfn_docgen.config import AppConfig, AppContext
from cfn_docgen.domain.model.cfn_document_generator import document_generator_factory
from cfn_docgen.domain.services.cfn_docgen_service import CfnDocgenService, CfnDocgenServiceCommandOutput
from cfn_docgen.entrypoints.serverless.model.lambda_model import CfnDocgenServerlessUnitsOfWork, S3NotificationEvent, ServerlessArguement

DEST_BUCKET_NAME=os.environ["DEST_BUCKET_NAME"]
DEST_BUCKET_PREFIX=os.environ["DEST_BUCKET_PREFIX"]
if DEST_BUCKET_PREFIX.startswith("/"):
    DEST_BUCKET_PREFIX = DEST_BUCKET_PREFIX[1:] # type: ignore

def lambda_handler(event:Mapping[str, Optional[Any]], context:Any) -> List[str]:
    outputs:List[CfnDocgenServiceCommandOutput] = []
    app_context = AppContext(log_level=logging.INFO)

    try:
        s3_notification_event = S3NotificationEvent(**event) # type: ignore
    except Exception:
        app_context.log_error("failed to parse event. received request event may not be a S3 Notification Event.")
        return []
        
    try:
        units_of_work = CfnDocgenServerlessUnitsOfWork(
            args=ServerlessArguement(
                format="markdown",
                sources=[r.s3 for r in s3_notification_event.Records],
                dest_bucket=DEST_BUCKET_NAME,
                dest_prefix=DEST_BUCKET_PREFIX,
            ),
            context=app_context,
        )
    except Exception:
        app_context.log_error("failed to setup pairs of template sources and document dests")
        return []
    
    try:
        service = CfnDocgenService(
            cfn_template_provider_facotry=template_provider_factory,
            cfn_document_generator_factory=document_generator_factory,
            cfn_document_storage_factory=document_storage_facotory,
            cfn_specification_repository=CfnSpecificationRepository(
                context=app_context,
                source_url=AppConfig.DEFAULT_SPECIFICATION_URL,
                loader=RemoteFileLoader(context=app_context),
                cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=app_context),
                recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
            ),
            context=app_context,
        )
    except Exception:
        app_context.log_error("failed to setup CfnDocgenService")
        return []

    for command_input in units_of_work.provide():
        try:
            app_context.log_debug(f"start cfn-docgen process for template source [{command_input.template_source.source}] and document dest [{command_input.document_dest.dest}]")
            result = service.main(command_input=command_input)
            app_context.log_info(f"successfully generate document [{result.document_dest}] from template [{command_input.template_source.source}]")
            outputs.append(result)
        except Exception:
            app_context.log_warning(f"failed to generate document [{command_input.document_dest.dest}] from template [{command_input.template_source.source}]")
            continue

    print(app_context.log_messages.as_string(logging.INFO))
    return [o.document_dest for o in outputs]


import logging
import os
import pytest
from cfn_docgen.config import AppContext
from cfn_docgen.domain.model.cfn_template import CfnTemplateDefinition, CfnTemplateSource
from cfn_docgen.adapters.cfn_template_provider import CfnTemplateProvider
from cfn_docgen.adapters.internal.file_loader import file_loader_factory

@pytest.fixture
def context():
    return AppContext(log_level=logging.DEBUG)

@pytest.fixture
def input_yaml_file():
    return os.path.join(
        os.path.dirname(__file__), "..", "..", "..", "..", "docs", "sample-template.yaml"
    )
@pytest.fixture
def input_json_file():
    return os.path.join(
        os.path.dirname(__file__), "..", "..", "..", "..", "docs", "sample-template.json"
    )

@pytest.mark.parametrize("input_file", [
    (
        os.path.join(
            os.path.dirname(__file__), "..", "..", "..", "..", "docs", "sample-template.yaml"
        )
    ),
    (
        os.path.join(
            os.path.dirname(__file__), "..", "..", "..", "..", "docs", "sample-template.json"
        )
    )
])
def test_load_yaml_template(input_file:str, context:AppContext):
    provider = CfnTemplateProvider(
        file_loader_factory=file_loader_factory,
        context=context,
    )
    template_definition = provider.load_template(CfnTemplateSource(
        source=input_file,
        context=context,
    ))

    assert isinstance(template_definition, CfnTemplateDefinition)



from __future__ import annotations
from dataclasses import dataclass
import re
from typing import Literal, Mapping, Optional
from pydantic import BaseModel

from cfn_docgen.config import AppContext

class CfnSpecificationProperty(BaseModel):
    Documentation: Optional[str] = None
    DuplicatesAllowed: Optional[bool] = None
    ItemType: Optional[str] = None
    PrimitiveItemType: Optional[str] = None
    PrimitiveType: Optional[str] = None
    Required: Optional[bool] = None
    Type: Optional[str] = None
    UpdateType: Optional[Literal["Conditional", "Immutable", "Mutable"]] = None

class CfnSpecificationResourceTypeAttribute(BaseModel):
    ItemType: Optional[str] = None
    PrimitiveItemType: Optional[str] = None
    PrimitiveType: Optional[str] = None
    Type: Optional[str] = None

class CfnSpecificationResourceType(BaseModel):
    AdditionalProperties: Optional[bool] = None
    Attributes: Mapping[str, CfnSpecificationResourceTypeAttribute] = {}
    Documentation: Optional[str] = None
    Properties: Mapping[str, CfnSpecificationProperty] = {}


class CfnSpecificationPropertyType(BaseModel):
    Documentation: Optional[str] = None
    DuplicatesAllowed: Optional[bool] = None
    ItemType: Optional[str] = None
    PrimitiveType: Optional[str] = None
    PrimitiveItemType: Optional[str] = None
    Required: Optional[bool] = None
    Type: Optional[str] = None
    UpdateType: Optional[Literal["Conditional", "Immutable", "Mutable"]] = None
    Properties: Optional[Mapping[str, CfnSpecificationProperty]] = None


@dataclass
class CfnSpecificationForResource:
    ResourceSpec: CfnSpecificationResourceType
    PropertySpecs: Mapping[str, CfnSpecificationPropertyType]

class CfnSpecification(BaseModel):
    ResourceSpecificationVersion: str
    ResourceTypes: Mapping[str, CfnSpecificationResourceType]
    PropertyTypes: Mapping[str, CfnSpecificationPropertyType]


class CfnSpecificationResourceTypeName:
    fullname:str
    __pattern = r"^[a-zA-Z0-9]+::[a-zA-Z0-9]+::[a-zA-Z0-9]+$"

    def __init__(self, resource_type:str, context:AppContext) -> None:
        try:
            assert re.match(self.__pattern, resource_type) is not None, f"Valid pattern for ResourceType is {self.__pattern} (e.g. AWS::EC2::Instance)"
        except AssertionError as ex:
            context.log_error(f"failed to instanciate for resource type {resource_type}. {str(ex)}")
            raise ex

        self.fullname = resource_type

class CfnSpecificationPropertyTypeName:
    fullname:str
    __pattern = r"^[a-zA-Z0-9]+::[a-zA-Z0-9]+::[a-zA-Z0-9]+\.[a-zA-Z0-9]+$|^Tag$"

    def __init__(self, property_type:str, context:AppContext) -> None:
        try:
            assert re.match(self.__pattern, property_type) is not None, f"Valid pattern for PropertyType is {self.__pattern} (e.g. AWS::EC2::Instance.BlockDeviceMapping)"
        except AssertionError as ex:
            context.log_error(f"failed to instanciate for property type {property_type}. {str(ex)}")
            raise ex
        self.fullname = property_type

    @property
    def prefix(self, ) -> str:
        return self.fullname.split(".")[0]
    
    @property
    def suffix(self, ) -> str:
        return self.fullname.split(".")[-1]


import logging
import os
import shutil
from typing import List
import boto3
import pytest
from cfn_docgen.adapters.internal.file_loader import file_loader_factory
from cfn_docgen.config import AppContext
from cfn_docgen.domain.model.cfn_document_generator import CfnDocumentDestination
from cfn_docgen.domain.model.cfn_template import CfnTemplateSource
from cfn_docgen.domain.services.cfn_docgen_service import CfnDocgenServiceCommandInput

from cfn_docgen.entrypoints.cli.model.cli_model import CfnDocgenCLIUnitsOfWork, CliArguement

INPUT_MASTER_FILE=os.path.join(
    os.path.dirname(__file__),
    "..", "..", "..", "..", "..", "..", "docs", "sample-template.yaml"
)
EXPECTED_MASTER_FILE=os.path.join(
    os.path.dirname(__file__),
    "..", "..", "..", "..", "..", "..", "docs", "sample-template.md"
)

# local directories and files
INPUT_ROOT_DIR=os.path.join(os.path.dirname(__file__), "data", "input")
INPUT_DIR1 = os.path.join(INPUT_ROOT_DIR, "dir1")
INPUT_FILE1 = os.path.join(INPUT_DIR1, "sample-template.yaml")
INPUT_DIR2 = os.path.join(INPUT_DIR1, "dir2")
INPUT_FILE2 = os.path.join(INPUT_DIR2, "sample-template.yaml")

OUTPUT_ROOT_DIR=os.path.join(os.path.dirname(__file__), "data", "output")
OUTPUT_DIR1 = os.path.join(OUTPUT_ROOT_DIR, "dir1")
OUTPUT_MD_FILE1 = os.path.join(OUTPUT_DIR1, "sample-template.md")
OUTPUT_DIR2 = os.path.join(OUTPUT_DIR1, "dir2")
OUTPUT_MD_FILE2 = os.path.join(OUTPUT_DIR2, "sample-template.md")


# s3 bucket and objects
s3_client = boto3.client("s3") # type: ignore
TEST_BUCKET_NAME=os.environ["TEST_BUCKET_NAME"]
INPUT_ROOT_PREFIX=f"s3://{TEST_BUCKET_NAME}/data/input"
INPUT_PREFIX1=INPUT_ROOT_PREFIX+"/dir1"
INPUT_KEY1=INPUT_PREFIX1+"/sample-template.yaml"
INPUT_PREFIX2=INPUT_PREFIX1+"/dir2"
INPUT_KEY2=INPUT_PREFIX2+"/sample-template.yaml"

OUTPUT_ROOT_PREFIX=f"s3://{TEST_BUCKET_NAME}/data/output"
OUTPUT_PREFIX1=OUTPUT_ROOT_PREFIX+"/dir1"
OUTPUT_MD_KEY1=OUTPUT_PREFIX1+"/sample-template.md"
OUTPUT_PREFIX2=OUTPUT_PREFIX1+"/dir2"
OUTPUT_MD_KEY2=OUTPUT_PREFIX2+"/sample-template.md"


@pytest.fixture(scope="class", autouse=True)
def class_local_dirs_and_files():
    # cleanup local dirs
    shutil.rmtree(INPUT_ROOT_DIR, ignore_errors=True)
    shutil.rmtree(OUTPUT_ROOT_DIR, ignore_errors=True)

    # prepare input files and dirs
    os.makedirs(INPUT_DIR2, exist_ok=True)
    shutil.copy(INPUT_MASTER_FILE, INPUT_FILE1)
    shutil.copy(INPUT_MASTER_FILE, INPUT_FILE2)


@pytest.fixture(scope="class", autouse=True)
def class_s3_bucket_and_keys():
    # cleanup s3 keys
    for prefix in [INPUT_ROOT_PREFIX, OUTPUT_ROOT_PREFIX]:
        res = s3_client.list_objects(
            Bucket=TEST_BUCKET_NAME,
            Prefix=prefix,
        )
        contents = res.get("Contents", None)
        if contents is None or len(contents) == 0: # type: ignore
            continue
        s3_client.delete_objects(
            Bucket=TEST_BUCKET_NAME,
            Delete={
                "Objects": [
                    {"Key": content["Key"]} for content in contents # type: ignore
                ]
            }
        )
    
    # prepare input keys
    for key in [INPUT_KEY1, INPUT_KEY2]:
        s3_client.upload_file(
            Bucket=TEST_BUCKET_NAME,
            Key=key,
            Filename=INPUT_MASTER_FILE
        )


@pytest.fixture(scope="function", autouse=True)
def function_local_dirs_and_files():
    # cleanup output dirs
    shutil.rmtree(OUTPUT_ROOT_DIR, ignore_errors=True)

@pytest.fixture(scope="function", autouse=True)
def function_s3_bucket_and_keys():
    # cleanup s3 keys
    res = s3_client.list_objects_v2(
        Bucket=TEST_BUCKET_NAME,
        Prefix=OUTPUT_ROOT_PREFIX,
    )
    contents = res.get("Contents", None)
    if contents is not None and len(contents) > 0: # type: ignore
        s3_client.delete_objects(
            Bucket=TEST_BUCKET_NAME,
            Delete={
                "Objects": [
                    {"Key": content["Key"]} for content in contents # type: ignore
                ]
            }
        )

@pytest.fixture
def context():
    return AppContext(
        log_level=logging.DEBUG,
    )

@pytest.mark.parametrize("case,args,expected", [
    (
        "local source file and local dest file",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_FILE1,
            dest=OUTPUT_MD_FILE1,
        ),
        [
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_FILE1, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_FILE1, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            )
        ]
    ),
    (
        "local source file and local dest dir",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_FILE1,
            dest=OUTPUT_DIR1,
        ),
        [
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_FILE1, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_FILE1, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            )
        ]
    ),
    (
        "local source file and s3 dest key",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_FILE1,
            dest=OUTPUT_MD_KEY1,
        ),
        [
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_FILE1, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_KEY1, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            )
        ]
    ),
    (
        "local source file and s3 dest prefix",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_FILE1,
            dest=OUTPUT_PREFIX1,
        ),
        [
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_FILE1, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_KEY1, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            )
        ]
    ),
    (
        "local source dir and local dest dir ",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_ROOT_DIR,
            dest=OUTPUT_ROOT_DIR,
        ),
        [
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_FILE2, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_FILE2, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            ),
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_FILE1, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_FILE1, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            ),
        ]
    ),
    (
        "local source dir and s3 dest prefix", 
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_ROOT_DIR,
            dest=OUTPUT_ROOT_PREFIX,
        ),
        [
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_FILE2, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_KEY2, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            ),
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_FILE1, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_KEY1, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            ),
        ]
    ),
    (
        "s3 source key and local dest file",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_KEY1,
            dest=OUTPUT_MD_FILE1,
        ),
        [
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_KEY1, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_FILE1, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            ),
        ]
    ),
    (
        "s3 source key and local dest dir",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_KEY1,
            dest=OUTPUT_DIR1,
        ),
        [
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_KEY1, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_FILE1, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            ),
        ]
    ),
    (
        "s3 source key and s3 dest key",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_KEY1,
            dest=OUTPUT_MD_KEY1,
        ),
        [
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_KEY1, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_KEY1, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            ),
        ]
    ),
    (
        "s3 source key and s3 dest prefix",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_KEY1,
            dest=OUTPUT_PREFIX1,
        ),
        [
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_KEY1, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_KEY1, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            ),
        ]
    ),
    (
        "s3 source prefix and local dest dir",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_ROOT_PREFIX,
            dest=OUTPUT_ROOT_DIR,
        ),
        [
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_KEY2, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_FILE2, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            ),
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_KEY1, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_FILE1, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            ),
        ]
    ),
    (
        "s3 source prefix and s3 dest prefix",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_ROOT_PREFIX,
            dest=OUTPUT_ROOT_PREFIX,
        ),
        [
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_KEY2, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_KEY2, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            ),
            CfnDocgenServiceCommandInput(
                template_source=CfnTemplateSource(INPUT_KEY1, context=AppContext(log_level=logging.DEBUG)),
                document_dest=CfnDocumentDestination(OUTPUT_MD_KEY1, context=AppContext(log_level=logging.DEBUG)),
                fmt="markdown",
            ),
        ]
    ),

])
def test_CfnDocgenCLIUnitsOfWork_build_units_of_work_valid_pattern(
    case:str, # type: ignore
    args:CliArguement,
    expected:List[CfnDocgenServiceCommandInput],
    context:AppContext,
):
    
    units_of_work = CfnDocgenCLIUnitsOfWork(
        args=args,
        file_loader_factory=file_loader_factory,
        context=context
    )

    assert len(units_of_work.units_of_work) == len(expected)
    for u, e in zip(units_of_work.provide(), expected):
        assert u.template_source.source == e.template_source.source
        assert u.document_dest.dest == e.document_dest.dest

@pytest.mark.parametrize("case,args", [
    (
        "local source dir and local dest file",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_DIR1,
            dest=OUTPUT_MD_FILE1,
        ),
    ),
    (
        "local source dir and s3 dest key",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_DIR1,
            dest=OUTPUT_MD_KEY1,
        ),
    ),
    (
        "s3 source prefix and local dest file",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_PREFIX1,
            dest=OUTPUT_MD_FILE1,
        ),
    ),
    (
        "s3 source prefix and s3 dest key",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_PREFIX1,
            dest=OUTPUT_MD_KEY1,
        ),
    ),
])
def test_CfnDocgenCLIUnitsOfWork_build_units_of_work_invalid_pattern(
    case:str,
    args:CliArguement,
    context:AppContext,
    caplog:pytest.LogCaptureFixture,
):
    caplog.set_level(logging.ERROR)
    
    with pytest.raises(
        AssertionError,
        match="no valid template sources and document dests are provided"
    ):
        CfnDocgenCLIUnitsOfWork(
            args=args,
            file_loader_factory=file_loader_factory,
            context=context
        )
    
    error_msg = f"invalid arguement pattern: source {args.source} is a directory, while dest {args.dest} is a single file"
    assert error_msg in context.log_messages.as_string(logging.ERROR)
    assert caplog.records[0].message == error_msg


@pytest.mark.parametrize("case,args", [
    (
        "not exist template source",
        CliArguement(
            subcommand="docgen",
            format="markdown",
            source="not-exist-source",
            dest=OUTPUT_MD_FILE1,
        ),
    ),
])
def test_CfnDocgenCLIUnitsOfWork_build_units_of_work_not_exist_source(
    case:str,
    args:CliArguement,
    context:AppContext,
    caplog:pytest.LogCaptureFixture,
):
    caplog.set_level(logging.ERROR)
    
    with pytest.raises(
        AssertionError,
        match="no valid template sources and document dests are provided"
    ):
        CfnDocgenCLIUnitsOfWork(
            args=args,
            file_loader_factory=file_loader_factory,
            context=context
        )
    
    error_msg =  f"failed to list template sources at [{args.source}]"
    assert error_msg in context.log_messages.as_string(logging.ERROR)
    assert caplog.records[0].message == error_msg and caplog.records[0].funcName == "build_units_of_work"





import hashlib
import os
from cfn_docgen.config import AppContext
from cfn_docgen.domain.ports.cache import IFileCache

HexHashString = str

class NoFileCache(IFileCache):

    def __init__(self, cache_root_dir: str, context: AppContext) -> None:
        super().__init__(cache_root_dir, context)
        
    def get(self, filepath: str) -> str | None:
        return None
    
    def put(self, filepath: str, body: str):
        pass

class LocalFileCache(IFileCache):

    def __init__(self, cache_root_dir: str, context: AppContext) -> None:
        super().__init__(cache_root_dir, context)
        os.makedirs(self.cache_root_dir, exist_ok=True)
        context.log_debug(f"create directory for cache at [{self.cache_root_dir}]")

        self.encoding = "UTF-8"

    def hash(self, string:str) -> HexHashString:
        h = hashlib.md5(string.encode())
        return h.hexdigest()
    
    def put(self, filepath: str, body: str):
        filepath_hash = self.hash(filepath)
        cache_filepath = os.path.join(self.cache_root_dir, filepath_hash)
        with open(cache_filepath, "w", encoding=self.encoding) as fp:
            fp.write(body)

    def get(self, filepath: str) -> str | None:
        filepath_hash = self.hash(filepath)
        cache_filepath = os.path.join(self.cache_root_dir, filepath_hash)
        if os.path.isfile(cache_filepath):
            self.context.log_debug(f"cache [{cache_filepath}] hit")
            with open(cache_filepath, "r", encoding=self.encoding) as fp:
                body = fp.read()
            return body
        return None


from abc import ABC, abstractmethod
from typing import Callable
from cfn_docgen.config import AppContext
from cfn_docgen.domain.model.cfn_template import CfnTemplateDefinition, CfnTemplateSource

from cfn_docgen.domain.ports.internal.file_loader import IFileLoader

class ICfnTemplateProvider(ABC):

    def __init__(
        self,
        file_loader_factory:Callable[[CfnTemplateSource, AppContext], IFileLoader],
        context:AppContext,
    ) -> None:
        self.context = context
        self.file_loader_factory = file_loader_factory

    @abstractmethod
    def load_template(self, template_source:CfnTemplateSource) -> CfnTemplateDefinition:
        """load cfn template file(either json or yaml format) as dict"""



from __future__ import annotations
import logging
import os
from typing import Any, Mapping, Optional, cast

import pytest
from cfn_docgen.adapters.cfn_specification_repository import CfnSpecificationRepository
from cfn_docgen.adapters.internal.cache import LocalFileCache
from cfn_docgen.adapters.internal.file_loader import  RemoteFileLoader
from cfn_docgen.config import AppConfig, AppContext
from cfn_docgen.domain.model.cfn_template import (
    CfnTemplateConditionsNode,
    CfnTemplateDefinition,
    CfnTemplateMappingsNode,
    CfnTemplateMetadataCfnDocgenDefinition,
    CfnTemplateMetadataDefinition,
    CfnTemplateMetadataInterface,
    CfnTemplateMetadataParameterGroup,
    CfnTemplateMetadataParameterGroupLabel,
    CfnTemplateOutputDefinition,
    CfnTemplateOutputsNode,
    CfnTemplateParameterDefinition,
    CfnTemplateParametersNode,
    CfnTemplateResourceDefinition,
    CfnTemplateResourceMetadataCfnDocgenDefinition,
    CfnTemplateResourceMetadataDefinition,
    CfnTemplateResourcesNode,
    CfnTemplateRuleAssertDefinition,
    CfnTemplateRuleDefinition,
    CfnTemplateRulesNode,
    CfnTemplateSource,
    CfnTemplateTree
)

@pytest.fixture
def context():
    return AppContext(log_level=logging.DEBUG)

def test_CfnTemplateMetadataDefinition_get_resource_description():
    description = "some-description"
    cfn_docgen = CfnTemplateResourceMetadataCfnDocgenDefinition(
        Description=description
    )

    d = cfn_docgen.get_resource_description()
    assert d is not None and d == description

@pytest.mark.parametrize("definition", [
    (CfnTemplateResourceMetadataCfnDocgenDefinition())

])
def CfnTemplateResourceMetadataCfnDocgenDefinition_get_resource_description_none(cfn_docgen:CfnTemplateResourceMetadataCfnDocgenDefinition):
    assert cfn_docgen.get_resource_description() is None

@pytest.mark.parametrize("json_path,expected,", [
    ("$.ImageId", "imageid"),
    ("$.CpuOptions.CoreCount", "corecount"),
    ("$.BlockDeviceMappings", None),
    ("$.BlockDeviceMappings[1].Ebs.VolumeType", "volumetype"),
    ("$.Tags", "tags"),
    ("$.NotExist", None),
])
def test_CfnTemplateResourceMetadataCfnDocgenDefinition_get_property_description_by_json_path(
    json_path:str, expected:Optional[str],
):
    definition = CfnTemplateResourceMetadataCfnDocgenDefinition(
        Properties={
            "ImageId": "imageid",
            "CpuOptions": {
                "CoreCount": "corecount"
            },
            "BlockDeviceMappings": [
                {},
                {
                    "Ebs": {
                        "VolumeType": "volumetype"
                    }
                }
            ],
            "Tags": "tags"
        }
    )
    d = definition.get_property_description_by_json_path(json_path)
    if expected is None:
        assert d is None
    else:
        assert d == expected





def test_CfnTemplateParametersNode(context:AppContext):
    parameter_definitions:Mapping[str, CfnTemplateParameterDefinition] = {
        "param11": CfnTemplateParameterDefinition(Type="String"),
        "param12": CfnTemplateParameterDefinition(Type="Number"),
        "param21": CfnTemplateParameterDefinition(Type="CommaDelimitedList"),
        "param": CfnTemplateParameterDefinition(Type="AWS::EC2::AvailabilityZone::Name"),
    }
    parameter_groups = [
        CfnTemplateMetadataParameterGroup(
            Label=CfnTemplateMetadataParameterGroupLabel(default="group1"),
            Parameters=["param11", "param12"],
        ),
        CfnTemplateMetadataParameterGroup(
            Label=CfnTemplateMetadataParameterGroupLabel(default="group2"),
            Parameters=["param21"],
        ),
        CfnTemplateMetadataParameterGroup(
            Label=CfnTemplateMetadataParameterGroupLabel(default="group3"),
            Parameters=["param3"],
        ),
    ]
    node = CfnTemplateParametersNode(
        definitions=parameter_definitions,
        parameter_groups=parameter_groups,
        context=context,
    )

    group_node1 = node.group_nodes.get("group1", None)
    group_node2 = node.group_nodes.get("group2", None)
    group_node3 = node.group_nodes.get("group3", None)
    independent_group_node = node.group_nodes.get(CfnTemplateParametersNode.group_name_for_independent_parameters, None)
    assert group_node1 is not None and len(group_node1.leaves.keys()) == 2
    assert group_node2 is not None and len(group_node2.leaves.keys()) == 1
    assert group_node3 is not None and len(group_node3.leaves.keys()) == 0
    assert independent_group_node is not None and len(independent_group_node.leaves.keys()) == 1

    parameter_leaf11 = group_node1.leaves.get("param11", None)
    parameter_leaf12 = group_node1.leaves.get("param12", None)
    parameter_leaf21 = group_node2.leaves.get("param21", None)
    independent_parameter_leaf = independent_group_node.leaves.get("param", None)
    assert parameter_leaf11 is not None and parameter_leaf11.definition.Type == "String"
    assert parameter_leaf12 is not None and parameter_leaf12.definition.Type == "Number"
    assert parameter_leaf21 is not None and parameter_leaf21.definition.Type == "CommaDelimitedList"
    assert independent_parameter_leaf is not None and independent_parameter_leaf.definition.Type == "AWS::EC2::AvailabilityZone::Name"

def test_CfnTemplateMappingsNode(context:AppContext):
    mappings_definition:Mapping[str, Mapping[str, Mapping[str, Any]]] = {
        "map1": {
            "name11": {
                "key111": "value111",
                "key112": "value112"
            },
            "name12": {
                "key121": "value121",
                "key122": "value122"
            }
        },
        "map2": {
            "name21": {
                "key211": "value211",
                "key212": "value212"
            },
            "name22": {
                "key221": "value221",
                "key222": "value222"
            }
        }

    }
    descriptions:Mapping[str, str] = {
        "map1": "description1",
        "map3": "description3"
    }
    node = CfnTemplateMappingsNode(
        definitions=mappings_definition,
        descriptions=descriptions,
        context=context,
    )

    leaf1 = node.mapping_leaves["map1"]
    assert leaf1.definition["name11"]["key111"] == "value111"
    assert leaf1.definition["name12"]["key122"] == "value122"
    assert leaf1.description is not None and leaf1.description == "description1"

    leaf2 = node.mapping_leaves["map2"]
    assert leaf2.definition["name21"]["key211"] == "value211"
    assert leaf2.definition["name22"]["key222"] == "value222"
    assert leaf2.description is None

    with pytest.raises(KeyError):
        node.mapping_leaves["map3"]
    

def test_CfnTemplateRulesNode(context:AppContext):
    rule_definitions:Mapping[str, CfnTemplateRuleDefinition] = {
        "rule1": CfnTemplateRuleDefinition(
            RuleCondition=None,
            Assertions=[
                CfnTemplateRuleAssertDefinition(
                    AssertDescription="assertdescription",
                    Assert={None: None}
                )
            ]
        ),
        "rule2": CfnTemplateRuleDefinition(
            RuleCondition={"rule": "condition"},
            Assertions=[
                CfnTemplateRuleAssertDefinition(
                    Assert={None: None},
                    AssertDescription="assertdescription"
                )
            ]
        )
    }
    descriptions = {
        "rule1": "description1",
        "rule3": "description3"
    }
    node = CfnTemplateRulesNode(
        definitions=rule_definitions,
        descriptions=descriptions,
        context=context,
    )

    leaf1 = node.rule_leaves["rule1"]
    assert leaf1.definition.RuleCondition is None
    assert leaf1.definition.Assertions[0].AssertDescription == "assertdescription"
    assert leaf1.description is not None and leaf1.description == "description1"

    leaf2 = node.rule_leaves["rule2"]
    assert leaf2.definition.RuleCondition is not None and leaf2.definition.RuleCondition["rule"] == "condition"
    assert leaf2.description is None

    with pytest.raises(KeyError):
        node.rule_leaves["rule3"]
    

def test_CfnTemplateConditionsNode(context:AppContext):
    definitions:Mapping[str, Mapping[str, Any]] = {
        "cond1": {
            "key1": "value1"
        },
        "cond2": {
            "key2": "value2"
        }
    }
    description = {
        "cond1": "description1",
        "cond3": "descirption3"
    }
    node = CfnTemplateConditionsNode(
        definitions=definitions,
        descriptions=description,
        context=context,
    )

    leaf1 = node.condition_leaves["cond1"]
    assert leaf1.definition["key1"] == "value1"
    assert leaf1.descirption is not None and leaf1.descirption == "description1"

    leaf2 = node.condition_leaves["cond2"]
    assert leaf2.definition["key2"] == "value2"
    assert leaf2.descirption is None

    with pytest.raises(KeyError):
        node.condition_leaves["cond3"]

def test_CfnTemplateOutputsNode(context:AppContext):
    definitions:Mapping[str, CfnTemplateOutputDefinition] = {
        "out1": CfnTemplateOutputDefinition(
            Description="description1",
            Value="value1"
        ),
        "out2": CfnTemplateOutputDefinition(
            Description=None,
            Value="value2"
        ),
    }

    node = CfnTemplateOutputsNode(
        definitions=definitions,
        context=context,
    )


    leaf1 = node.output_leaves.get("out1")
    assert leaf1 is not None
    leaf2 = node.output_leaves.get("out2")
    assert leaf2 is not None

@pytest.fixture
def spec_repository(context:AppContext):
    return CfnSpecificationRepository(
        context=context,
        source_url=AppConfig.DEFAULT_SPECIFICATION_URL,
        loader=RemoteFileLoader(context=context),
        cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=context),
        recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
    )

def all_resource_types():
    context = AppContext(log_level=logging.DEBUG)
    spec_repository = CfnSpecificationRepository(
        context=context,
        source_url=AppConfig.DEFAULT_SPECIFICATION_URL,
        loader=RemoteFileLoader(context=context),
        cache=LocalFileCache(AppConfig.CACHE_ROOT_DIR, context=context),
        recursive_resource_types=AppConfig.RECURSIVE_RESOURCE_TYPES,
    )
    return list(spec_repository.spec.ResourceTypes.keys())

def test_CfnTemplateResourcesNode_aws_ec2_instance(
    spec_repository:CfnSpecificationRepository,
    context:AppContext,
):
    resource_description = "resource-descirption"
    property_descriptions = {
        "ImageId": "imageid",
        "BlockDeviceMappings": [
            {
                "Ebs": "ebs"
            }
        ],
        "SecurityGroupIds": "securitygroupids",
        "Tags": [
            {},
            {
                "Key": "key"
            }
        ]
    }
    definitions:Mapping[str, CfnTemplateResourceDefinition] = {
        "Instance": CfnTemplateResourceDefinition(
            Type="AWS::EC2::Instance",
            DependsOn=["VPC"],
            DeletionPolicy="Retain",
            Metadata=CfnTemplateResourceMetadataDefinition(
                CfnDocgen=CfnTemplateResourceMetadataCfnDocgenDefinition(
                    Description=resource_description,
                    Properties=property_descriptions,
                )
            ),
            Properties={
                "ImageId": "IMAGEID",
                "BlockDeviceMappings": [
                    {
                        "Ebs": {
                            "Encrypted" : True,
                            "VolumeSize" : 0,
                            "VolumeType" : "VOLUMETYPE"
                        }
                    },
                    {
                        "NoDevice": {} 
                    }
                ],
                "CpuOptions": {
                    "CoreCount": 0
                },
                "SecurityGroupIds": [
                    "SECURITYGROUPID1",
                    "SECURITYGROUPID2",
                    "SECURITYGROUPID3",
                ],
                "Tags": [
                    {
                        "Key": "KEY1",
                        "Value": "VALUE1"
                    },
                    {
                        "Key": "KEY2",
                        "Value": "VALUE2"
                    }
                ]
            }
        )
    }

    resources_node = CfnTemplateResourcesNode(
        definitions=definitions,
        spec_repository=spec_repository,
        context=context,
    )

    instance_node = resources_node.resource_nodes.get("Instance")
    assert instance_node is not None
    assert instance_node.description is not None and instance_node.description == resource_description
    assert len(instance_node.depends_on) == 1 and instance_node.depends_on[0] == "VPC"
    assert instance_node.deletion_policy == "Retain"
    assert instance_node.update_policy is None
    assert instance_node.creation_policy is None
    assert instance_node.update_replace_policy == "Delete"
    assert instance_node.spec.Documentation is not None and instance_node.spec.Documentation == "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-instance.html"

    properties_node = instance_node.properties_node

    image_id_leaf = properties_node.property_leaves.get("ImageId")
    assert image_id_leaf is not None
    assert image_id_leaf.definition == "IMAGEID"
    assert image_id_leaf.json_path == "$.ImageId"
    assert image_id_leaf.description == "imageid"
    assert image_id_leaf.spec.PrimitiveType is not None and image_id_leaf.spec.PrimitiveType == "String"

    cpu_options_node = properties_node.property_nodes.get("CpuOptions")
    assert cpu_options_node is not None
    assert cpu_options_node.description is None
    assert cpu_options_node.json_path == "$.CpuOptions"
    assert cpu_options_node.spec is not None
    assert cpu_options_node.spec.Type is not None and cpu_options_node.spec.Type == "CpuOptions"
    core_count_leaf = cpu_options_node.property_leaves.get("CoreCount")
    assert core_count_leaf is not None
    assert core_count_leaf.description is None
    assert core_count_leaf.json_path == "$.CpuOptions.CoreCount"
    assert core_count_leaf.spec is not None
    assert core_count_leaf.spec.PrimitiveType is not None and core_count_leaf.spec.PrimitiveType == "Integer"
    assert core_count_leaf.definition is not None and core_count_leaf.definition == 0
    threads_per_core_leaf = cpu_options_node.property_leaves.get("ThreadsPerCore")
    assert threads_per_core_leaf is not None
    assert threads_per_core_leaf.description is None
    assert threads_per_core_leaf.json_path == "$.CpuOptions.ThreadsPerCore"
    assert threads_per_core_leaf.spec is not None
    assert threads_per_core_leaf.spec.PrimitiveType is not None and threads_per_core_leaf.spec.PrimitiveType == "Integer"
    assert threads_per_core_leaf.definition is None

    bdm_nodes = properties_node.property_nodes_list.get("BlockDeviceMappings")
    assert bdm_nodes is not None and len(bdm_nodes) == 2
    first_bdm_node = bdm_nodes[0]
    assert first_bdm_node.description is None
    assert first_bdm_node.json_path == "$.BlockDeviceMappings[0]"
    assert first_bdm_node.spec is not None
    assert first_bdm_node.spec.Type is not None and first_bdm_node.spec.Type == "List"
    assert first_bdm_node.spec.ItemType is not None and first_bdm_node.spec.ItemType == "BlockDeviceMapping"
    ebs_node = first_bdm_node.property_nodes.get("Ebs")
    assert ebs_node is not None
    assert ebs_node.description is not None and ebs_node.description == "ebs"
    assert ebs_node.json_path == "$.BlockDeviceMappings[0].Ebs"
    assert ebs_node.spec is not None
    assert ebs_node.spec.Type is not None and ebs_node.spec.Type == "Ebs"
    encrypted_leaf = ebs_node.property_leaves.get("Encrypted")
    assert encrypted_leaf is not None
    assert encrypted_leaf.description is None
    assert encrypted_leaf.json_path == "$.BlockDeviceMappings[0].Ebs.Encrypted"
    assert encrypted_leaf.spec is not None
    assert encrypted_leaf.spec.PrimitiveType is not None and encrypted_leaf.spec.PrimitiveType == "Boolean"
    assert encrypted_leaf.definition is not None and encrypted_leaf.definition == True
    second_bdm_node = bdm_nodes[1]
    assert second_bdm_node.description is None
    assert second_bdm_node.json_path == "$.BlockDeviceMappings[1]"
    assert second_bdm_node.spec is not None
    assert second_bdm_node.spec.Type is not None and second_bdm_node.spec.Type == "List"
    assert second_bdm_node.spec.ItemType is not None and second_bdm_node.spec.ItemType == "BlockDeviceMapping"
    nodevice_node = second_bdm_node.property_nodes.get("NoDevice")
    assert nodevice_node is not None
    assert nodevice_node.description is None
    assert nodevice_node.json_path == "$.BlockDeviceMappings[1].NoDevice"
    assert nodevice_node.spec is not None
    assert nodevice_node.spec.Type is not None and nodevice_node.spec.Type == "NoDevice"
    assert len(nodevice_node.property_leaves) == 0
    assert len(nodevice_node.property_nodes) == 0
    assert len(nodevice_node.property_nodes_list) == 0
    assert len(nodevice_node.property_nodes_map) == 0
    device_name_leaf = second_bdm_node.property_leaves.get("DeviceName")
    assert device_name_leaf is not None
    assert device_name_leaf.description is None
    assert device_name_leaf.json_path == "$.BlockDeviceMappings[1].DeviceName"
    assert device_name_leaf.spec is not None
    assert device_name_leaf.spec.PrimitiveType is not None and device_name_leaf.spec.PrimitiveType == "String"
    assert device_name_leaf.definition is None

    sg_ids_leaf = properties_node.property_leaves.get("SecurityGroupIds")
    assert sg_ids_leaf is not None
    assert sg_ids_leaf.description is not None and sg_ids_leaf.description == "securitygroupids"
    assert sg_ids_leaf.json_path == "$.SecurityGroupIds"
    assert sg_ids_leaf.spec is not None
    assert sg_ids_leaf.spec.Type is not None and sg_ids_leaf.spec.Type == "List"
    assert sg_ids_leaf.spec.PrimitiveItemType is not None and sg_ids_leaf.spec.PrimitiveItemType == "String"
    assert isinstance(sg_ids_leaf.definition, list)
    assert all([r == e for r, e in zip(sg_ids_leaf.definition, ["SECURITYGROUPID1", "SECURITYGROUPID2"])])

    tags_nodes = properties_node.property_nodes_list.get("Tags")
    assert tags_nodes is not None and len(tags_nodes) == 2, properties_node.property_nodes_list
    first_tag_node = tags_nodes[0]
    assert first_tag_node.description is None
    assert first_tag_node.json_path == "$.Tags[0]"
    assert first_tag_node.spec is not None
    assert first_tag_node.spec.Type is not None and first_tag_node.spec.Type == "List"
    assert first_tag_node.spec.ItemType is not None and first_tag_node.spec.ItemType == "Tag"
    first_key_leaf = first_tag_node.property_leaves.get("Key")
    assert first_key_leaf is not None
    assert first_key_leaf.description is None
    assert first_key_leaf.json_path == "$.Tags[0].Key"
    assert first_key_leaf.spec is not None
    assert first_key_leaf.spec.PrimitiveType is not None and first_key_leaf.spec.PrimitiveType == "String"
    assert first_key_leaf.definition is not None and first_key_leaf.definition == "KEY1"
    first_value_leaf = first_tag_node.property_leaves.get("Value")
    assert first_value_leaf is not None
    assert first_value_leaf.description is None
    assert first_value_leaf.json_path == "$.Tags[0].Value"
    assert first_value_leaf.spec is not None
    assert first_value_leaf.spec.PrimitiveType is not None and first_value_leaf.spec.PrimitiveType == "String"
    assert first_value_leaf.definition is not None and first_value_leaf.definition == "VALUE1"
    second_tag_node = tags_nodes[1]
    assert second_tag_node.description is None
    assert second_tag_node.json_path == "$.Tags[1]"
    assert second_tag_node.spec is not None
    assert second_tag_node.spec.Type is not None and first_tag_node.spec.Type == "List"
    assert second_tag_node.spec.ItemType is not None and first_tag_node.spec.ItemType == "Tag"
    second_key_leaf = second_tag_node.property_leaves.get("Key")
    assert second_key_leaf is not None
    assert second_key_leaf.description is not None and second_key_leaf.description == "key"
    assert second_key_leaf.json_path == "$.Tags[1].Key"
    assert second_key_leaf.spec is not None
    assert second_key_leaf.spec.PrimitiveType is not None and second_key_leaf.spec.PrimitiveType == "String"
    assert second_key_leaf.definition is not None and second_key_leaf.definition == "KEY2"
    second_value_leaf = second_tag_node.property_leaves.get("Value")
    assert second_value_leaf is not None
    assert second_value_leaf.description is None
    assert second_value_leaf.json_path == "$.Tags[1].Value"
    assert second_value_leaf.spec is not None
    assert second_value_leaf.spec.PrimitiveType is not None and first_value_leaf.spec.PrimitiveType == "String"
    assert second_value_leaf.definition is not None and second_value_leaf.definition == "VALUE2"
    

    key_name_node = properties_node.property_leaves.get("KeyName")
    assert key_name_node is not None
    assert key_name_node.description is None
    assert key_name_node.json_path == "$.KeyName"
    assert key_name_node.spec is not None
    assert key_name_node.spec.PrimitiveType is not None and key_name_node.spec.PrimitiveType == "String"
    assert key_name_node.definition is None



@pytest.mark.skipif(condition=os.environ.get("SKIP_LONG_TEST", False), reason="this test takes long time to run")
@pytest.mark.parametrize("resource_type", [
    (r) for r in all_resource_types()
])
def test_CfnTemplateResourcesNode_all_resource_types(
    resource_type:str, spec_repository:CfnSpecificationRepository,
    context:AppContext,
):
    definitions:Mapping[str, CfnTemplateResourceDefinition] = {
        resource_type: CfnTemplateResourceDefinition(
            Type=resource_type,
            Properties={}
        )
    }

    resources_node = CfnTemplateResourcesNode(
        definitions=definitions,
        spec_repository=spec_repository,
        context=context,
    )

    resource_node = resources_node.resource_nodes.get(resource_type)
    assert resource_node is not None
    assert resource_node.spec is not None
    properties_node = resource_node.properties_node
    assert (
        len(properties_node.property_leaves) > 0
        or len(properties_node.property_nodes) > 0
        or len(properties_node.property_nodes_list) > 0
        or len(properties_node.property_nodes_map)> 0
        # resource types without any properties
        or resource_type == "AWS::CloudFormation::WaitConditionHandle"
        or resource_type == "AWS::DevOpsGuru::LogAnomalyDetectionIntegration"
    )


def test_CfnTemplateResourcesNode_avoid_recursion_error(
    spec_repository:CfnSpecificationRepository,
    context:AppContext,
):
    definitions:Mapping[str, CfnTemplateResourceDefinition] = {
        "WithDefinition": CfnTemplateResourceDefinition(
            Type="AWS::WAFv2::RuleGroup",
            Properties={
                "Rules": [
                    {
                        "Statement": {
                            "AndStatement": {
                                "Statements": [
                                    {
                                        "RateBasedStatement": {
                                            "Limit": 500
                                        },
                                    },
                                    {
                                        "ByteMatchStatement": {
                                            "FieldToMatch": {
                                                "Method": {"Name": "GET"}
                                            }
                                        }
                                    }
                                ]
                            }
                        }
                    }
                ]
            },
        ),
        "WithoutDefinition": CfnTemplateResourceDefinition(
            Type="AWS::WAFv2::RuleGroup",
            Properties={},
        )
    }

    resources_node = CfnTemplateResourcesNode(
        definitions=definitions,
        spec_repository=spec_repository,
        context=context,
    )
    with_definition_node = resources_node.resource_nodes.get("WithDefinition")
    assert with_definition_node is not None
    rules = with_definition_node.properties_node.property_nodes_list.get("Rules")
    assert rules is not None and len(rules) == 1
    statement1 = rules[0].property_nodes.get("Statement")
    assert statement1 is not None
    and_statement = statement1.property_nodes.get("AndStatement")
    assert and_statement is not None
    statements = and_statement.property_nodes_list.get("Statements")
    assert statements is not None and len(statements) == 2
    byte_statement = statements[1].property_nodes.get("ByteMatchStatement")
    assert byte_statement is not None
    field_to_match = byte_statement.property_nodes.get("FieldToMatch")
    assert field_to_match is not None
    method = field_to_match.property_leaves.get("Method")
    assert method is not None and method.definition is not None and cast(Mapping[str, Any], method.definition)["Name"] == "GET"

    without_definition_node = resources_node.resource_nodes.get("WithoutDefinition")
    assert without_definition_node is not None
    rules_node = without_definition_node.properties_node.property_nodes_list["Rules"]
    assert len(rules_node) == 0


def test_CfnTemplateTree(
    spec_repository:CfnSpecificationRepository,
    context:AppContext,
):
    definition = CfnTemplateDefinition(
        AWSTemplateFormatVersion="2012-10",
        Description="template-description",
        Metadata=CfnTemplateMetadataDefinition(**{
            'AWS::CloudFormation::Interface': CfnTemplateMetadataInterface(
                ParameterGroups=[
                    CfnTemplateMetadataParameterGroup(
                        Label=CfnTemplateMetadataParameterGroupLabel(
                            default="group1"
                        ),
                        Parameters=["param1"]
                    )
                ]
            ),
            "CfnDocgen": CfnTemplateMetadataCfnDocgenDefinition(
                Description="cfn-docgen-template-description",
                Mappings={
                    "map1": "map1-description"
                },
                Rules={
                    "rule1": "rule1-description"
                },
                Conditions={
                    "cond1": "cond1-description"
                },
            )
        }), # type: ignore
        Parameters={
            "param1": CfnTemplateParameterDefinition(
                Type="String"
            ),
            "param2": CfnTemplateParameterDefinition(
                Type="Number"
            )
        },
        Mappings={
            "map1": {
                "name1": {
                    "key1": "value1"
                }
            }
        },
        Rules={
            "rule1": CfnTemplateRuleDefinition(
                Assertions=[
                    CfnTemplateRuleAssertDefinition(
                        Assert={
                            "assert1": "value1"
                        },
                        AssertDescription="assert1-description"
                    )
                ]
            )
        },
        Conditions={
            "cond1": {
                "key1": "value1"
            }
        },
        Resources={
            "resource1": CfnTemplateResourceDefinition(
                Metadata=CfnTemplateResourceMetadataDefinition(
                    CfnDocgen=CfnTemplateResourceMetadataCfnDocgenDefinition(
                        Description="resource1-description",
                        Properties={
                            "BlockDeviceMappings": [
                                {
                                    "Ebs": {
                                        "Encrypted": "encrypted-description"
                                    }
                                }
                            ]
                        },
                    )
                ),
                Type="AWS::EC2::Instance",
                Properties={
                    "BlockDeviceMappings": [
                        {
                            "Ebs": {
                                "Encrypted": True
                            }
                        }
                    ],
                    "ImageId": {"Ref": "param1"}
                }
            )
        },
        Outputs={
            "out1": CfnTemplateOutputDefinition(
                Value={"Ref": "resource1"}
            )
        }
    )

    tree = CfnTemplateTree(
        template_source=CfnTemplateSource("/tmp/sample-template.yaml", context=context),
        definition=definition,
        spec_repository=spec_repository,
        context=context,
    )
    
    assert tree.description == "template-description"
    assert tree.cfn_docgen_description == "cfn-docgen-template-description"


    param1_leaf = tree.parameters_node.group_nodes["group1"].leaves["param1"]
    assert param1_leaf.definition.Type == "String"
    param2_leaf = tree.parameters_node.group_nodes[CfnTemplateParametersNode.group_name_for_independent_parameters].leaves["param2"]
    assert param2_leaf.definition.Type == "Number"

    map1_leaf = tree.mappings_node.mapping_leaves["map1"]
    assert map1_leaf.description == "map1-description"

    rule1_leaf = tree.rules_node.rule_leaves["rule1"]
    assert rule1_leaf.description == "rule1-description"

    cond1_leaf = tree.conditions_node.condition_leaves["cond1"]
    assert cond1_leaf.descirption == "cond1-description"

    resource1_node = tree.resources_node.resource_nodes["resource1"]
    assert resource1_node.description == "resource1-description"
    image_id_leaf = resource1_node.properties_node.property_leaves["ImageId"]
    assert image_id_leaf.description is None
    assert image_id_leaf.definition is not None and cast(Mapping[str, str], image_id_leaf.definition)["Ref"] == "param1"
    bdm_node = resource1_node.properties_node.property_nodes_list["BlockDeviceMappings"][0]
    ebs_node = bdm_node.property_nodes["Ebs"]
    encrypted_leaf = ebs_node.property_leaves["Encrypted"]
    assert encrypted_leaf.definition == True
    assert encrypted_leaf.description is not None and encrypted_leaf.description == "encrypted-description"

    out1_leaf = tree.outputs_node.output_leaves["out1"]
    assert out1_leaf.definition.Value["Ref"] == "resource1"

@pytest.mark.parametrize("filepath", [
    ("../template.yaml"), ("/tmp/template.yaml"), ("template.yaml")
])
def test_CfnTemplateSource_localfilepath_basename(
    filepath:str,
    context:AppContext,
):
    source = CfnTemplateSource(filepath, context=context)
    assert source.basename == "template.yaml"
    




from __future__ import annotations
from abc import ABC, abstractmethod
from typing import List

from cfn_docgen.config import AppContext

class IFileLoader(ABC):

    def __init__(self, context:AppContext) -> None:
        self.context = context

    @abstractmethod
    def download(self, source:str) -> bytes:
        pass

    @abstractmethod
    def upload(self, body:bytes, dest:str) -> None:
        pass

    @abstractmethod
    def list(self, source:str) -> List[str]:
        pass


from abc import ABC, abstractmethod
from typing import List, Mapping
from cfn_docgen.config import AppContext
from cfn_docgen.domain.ports.cache import IFileCache
from cfn_docgen.domain.ports.internal.file_loader import IFileLoader

from cfn_docgen.domain.model.cfn_specification import CfnSpecificationPropertyTypeName, CfnSpecificationResourceTypeName, CfnSpecificationPropertyType, CfnSpecificationResourceType, CfnSpecificationForResource


class ICfnSpecificationRepository(ABC):

    def __init__(
        self,
        source_url:str,
        loader:IFileLoader, 
        cache:IFileCache, 
        recursive_resource_types:List[str],
        context:AppContext,
    ) -> None:
        self.context = context
        self.recursive_resource_types = recursive_resource_types
        super().__init__()

    @abstractmethod
    def get_resource_spec(self, resource_type:CfnSpecificationResourceTypeName) -> CfnSpecificationResourceType:
        pass

    @abstractmethod
    def get_property_spec(self, property_type:CfnSpecificationPropertyTypeName) -> CfnSpecificationPropertyType:
        pass


    @abstractmethod
    def list_properties_for_resource(self, resource_type:CfnSpecificationResourceTypeName) -> Mapping[str, CfnSpecificationPropertyType]:
        pass


    @abstractmethod
    def get_specs_for_resource(self, resource_type:CfnSpecificationResourceTypeName) -> CfnSpecificationForResource:
        pass

    @abstractmethod
    def is_recursive(self, resource_type:CfnSpecificationResourceTypeName) -> bool:
        pass




import os
from setuptools import find_packages, setup
from cfn_docgen import __version__
with open(os.path.join(
    os.path.dirname(__file__), "README.md"), "r", encoding="utf-8",
) as fp:
    long_description = fp.read()
with open(os.path.join(
    os.path.dirname(__file__), "requirements.txt"), "r", encoding="utf-8"
) as fp:
    requirements = fp.read()

setup(
    name = 'cfn-docgen',
    version = __version__,
    author = 'Takehiro Horie',
    author_email = 'horie.takehiro@outlook.jp',
    license = 'MIT License',
    description = 'Document generator from cfn template files.',
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = 'https://github.com/horietakehiro/cfn-docgen',
    package_dir={"": "src"},
    packages = find_packages(where="src", ),
    install_requires = [requirements],
    python_requires='>=3.10',
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
    ],
    entry_points = '''
        [console_scripts]
        cfn-docgen=cfn_docgen.entrypoints.cli.main:main
    '''
)


# pylint: disable=W0621
from typing import Callable
from cfn_docgen.adapters.internal.file_loader import document_loader_factory
from cfn_docgen.config import AppContext
from cfn_docgen.domain.model.cfn_document_generator import CfnDocumentDestination
from cfn_docgen.domain.ports.cfn_document_storage import ICfnDocumentStorage
from cfn_docgen.domain.ports.internal.file_loader import IFileLoader

def document_storage_facotory(dest:CfnDocumentDestination, context:AppContext) -> ICfnDocumentStorage:
    context.log_debug(f"return CfnDocumentStorage for dest [{dest}]")
    return CfnDocumentStorage(
        document_loader_factory=document_loader_factory,
        context=context,
    )

class CfnDocumentStorage(ICfnDocumentStorage):
    def __init__(
        self, 
        document_loader_factory: Callable[[CfnDocumentDestination, AppContext],IFileLoader],
        context: AppContext,
    ) -> None:
        super().__init__(document_loader_factory, context)

    def save_document(self, body: bytes, document_dest: CfnDocumentDestination) -> None:
        loader = self.document_loader_factory(document_dest, self.context)
        loader.upload(body, document_dest.dest)
        self.context.log_debug(f"save document to [{document_dest.dest}]")


import pytest
import os
import shutil
import boto3
from click.testing import CliRunner

from cfn_docgen.entrypoints.cli.main import main
from cfn_docgen.entrypoints.cli.model.cli_model import CliArguement

INPUT_MASTER_FILE=os.path.join(
    os.path.dirname(__file__),
    "..", "..", "..", "..", "..", "docs", "sample-template.yaml"
)
EXPECTED_MASTER_FILE=os.path.join(
    os.path.dirname(__file__),
    "..", "..", "..", "..", "..", "docs", "sample-template.md"
)

# local directories and files
INPUT_ROOT_DIR=os.path.join(os.path.dirname(__file__), "data", "input")
INPUT_DIR1 = os.path.join(INPUT_ROOT_DIR, "dir1")
INPUT_FILE1 = os.path.join(INPUT_DIR1, "sample-template.yaml")
INPUT_DIR2 = os.path.join(INPUT_DIR1, "dir2")
INPUT_FILE2 = os.path.join(INPUT_DIR2, "sample-template.yaml")

OUTPUT_ROOT_DIR=os.path.join(os.path.dirname(__file__), "data", "output")
OUTPUT_DIR1 = os.path.join(OUTPUT_ROOT_DIR, "dir1")
OUTPUT_MD_FILE1 = os.path.join(OUTPUT_DIR1, "sample-template.md")
OUTPUT_DIR2 = os.path.join(OUTPUT_DIR1, "dir2")
OUTPUT_MD_FILE2 = os.path.join(OUTPUT_DIR2, "sample-template.md")


# s3 bucket and objects
s3_client = boto3.client("s3") # type: ignore
TEST_BUCKET_NAME=os.environ["TEST_BUCKET_NAME"]
INPUT_ROOT_PREFIX=f"s3://{TEST_BUCKET_NAME}/data/input"
INPUT_PREFIX1=INPUT_ROOT_PREFIX+"/dir1"
INPUT_KEY1=INPUT_PREFIX1+"/sample-template.yaml"
INPUT_PREFIX2=INPUT_PREFIX1+"/dir2"
INPUT_KEY2=INPUT_PREFIX2+"/sample-template.yaml"

OUTPUT_ROOT_PREFIX=f"s3://{TEST_BUCKET_NAME}/data/output"
OUTPUT_PREFIX1=OUTPUT_ROOT_PREFIX+"/dir1"
OUTPUT_MD_KEY1=OUTPUT_PREFIX1+"/sample-template.md"
OUTPUT_PREFIX2=OUTPUT_PREFIX1+"/dir2"
OUTPUT_MD_KEY2=OUTPUT_PREFIX2+"/sample-template.md"


@pytest.fixture(scope="class", autouse=True)
def class_local_dirs_and_files():
    # cleanup local dirs
    shutil.rmtree(INPUT_ROOT_DIR, ignore_errors=True)
    shutil.rmtree(OUTPUT_ROOT_DIR, ignore_errors=True)

    # prepare input files and dirs
    os.makedirs(INPUT_DIR2, exist_ok=True)
    shutil.copy(INPUT_MASTER_FILE, INPUT_FILE1)
    shutil.copy(INPUT_MASTER_FILE, INPUT_FILE2)


@pytest.fixture(scope="class", autouse=True)
def class_s3_bucket_and_keys():
    # cleanup s3 keys
    for prefix in [INPUT_ROOT_PREFIX, OUTPUT_ROOT_PREFIX]:
        res = s3_client.list_objects(
            Bucket=TEST_BUCKET_NAME,
            Prefix=prefix,
        )
        contents = res.get("Contents", None)
        if contents is None or len(contents) == 0: # type: ignore
            continue
        s3_client.delete_objects(
            Bucket=TEST_BUCKET_NAME,
            Delete={
                "Objects": [
                    {"Key": content["Key"]} for content in contents # type: ignore
                ]
            }
        )
    
    # prepare input keys
    for key in [INPUT_KEY1, INPUT_KEY2]:
        s3_client.upload_file(
            Bucket=TEST_BUCKET_NAME,
            Key=key,
            Filename=INPUT_MASTER_FILE
        )


@pytest.fixture(scope="function", autouse=True)
def function_local_dirs_and_files():
    # cleanup output dirs
    shutil.rmtree(OUTPUT_ROOT_DIR, ignore_errors=True)

@pytest.fixture(scope="function", autouse=True)
def function_s3_bucket_and_keys():
    # cleanup s3 keys
    res = s3_client.list_objects_v2(
        Bucket=TEST_BUCKET_NAME,
        Prefix=OUTPUT_ROOT_PREFIX,
    )
    contents = res.get("Contents", None)
    if contents is not None and len(contents) > 0: # type: ignore
        s3_client.delete_objects(
            Bucket=TEST_BUCKET_NAME,
            Delete={
                "Objects": [
                    {"Key": content["Key"]} for content in contents # type: ignore
                ]
            }
        )


def test_cli_local_file_source_local_file_dest():
    args = CliArguement(
        subcommand="docgen",
        format="markdown",
        source=INPUT_FILE1,
        dest=OUTPUT_MD_FILE1,
    )

    runner = CliRunner()
    result = runner.invoke(main, args=args.as_list())
    assert result.exit_code == 0

    with open(OUTPUT_MD_FILE1, "rb") as fp:
        output = fp.read()
    with open(EXPECTED_MASTER_FILE, "rb") as fp:
        example = fp.read()
    
    assert output == example


def test_cli_s3_key_source_s3_key_dest():
    args = CliArguement(
        subcommand="docgen",
        format="markdown",
        source=f"s3://{TEST_BUCKET_NAME}/{INPUT_KEY1}",
        dest=f"s3://{TEST_BUCKET_NAME}/{OUTPUT_MD_KEY1}",
    )
    runner = CliRunner()
    result = runner.invoke(main, args=args.as_list())
    assert result.exit_code == 0

    with open(EXPECTED_MASTER_FILE, "rb") as fp:
        expected = fp.read()
    res = s3_client.get_object(
        Bucket=TEST_BUCKET_NAME, Key=OUTPUT_MD_KEY1,
    )
    assert res["Body"].read() == expected


def test_cli_local_folder_source_local_folder_dest():
    args = CliArguement(
        subcommand="docgen",
        format="markdown",
        source=INPUT_ROOT_DIR,
        dest=OUTPUT_ROOT_DIR,
    )

    runner = CliRunner()
    result = runner.invoke(main, args=args.as_list())
    assert result.exit_code == 0

    with open(EXPECTED_MASTER_FILE, "r") as fp:
        expected = fp.read()
    
    for output_file in [OUTPUT_MD_FILE1, OUTPUT_MD_FILE2]:
        with open(output_file, "r") as fp:
            assert fp.read() == expected

def test_cli_s3_prefix_source_s3_prefix_dest():
    args = CliArguement(
        subcommand="docgen",
        format="markdown",
        source=f"s3://{TEST_BUCKET_NAME}/{INPUT_ROOT_PREFIX}",
        dest=f"s3://{TEST_BUCKET_NAME}/{OUTPUT_ROOT_PREFIX}",
    )
    runner = CliRunner()
    result = runner.invoke(main, args=args.as_list())
    assert result.exit_code == 0

    with open(EXPECTED_MASTER_FILE, "rb") as fp:
        expected = fp.read()
    
    for output_key in [OUTPUT_MD_KEY1, OUTPUT_MD_KEY2]:
        res = s3_client.get_object(Bucket=TEST_BUCKET_NAME, Key=output_key)
        assert res["Body"].read() == expected

def test_cli_build_units_of_work_fail():
    args = CliArguement(
        subcommand="docgen",
        format="markdown",
        source="not-exist-source",
        dest=f"s3://{TEST_BUCKET_NAME}/{OUTPUT_ROOT_PREFIX}",
    )
    runner = CliRunner()
    result = runner.invoke(main, args=args.as_list())
    assert result.exit_code == 1
    assert "[ERROR] failed to setup pairs of template sources and document dests" in result.output


def test_cli_continue():
    
    # temporary replace template source with invalid one
    try:
        with open(INPUT_FILE1, "wb") as fp:
            fp.write(b"invalid")
        args = CliArguement(
            subcommand="docgen",
            format="markdown",
            source=INPUT_ROOT_DIR,
            dest=OUTPUT_ROOT_DIR,
            debug=False,
        )
        runner = CliRunner()
        result = runner.invoke(main, args=args.as_list())
        assert result.exit_code == 0
        assert f"[WARNING] failed to generate document [{OUTPUT_MD_FILE1}] from template [{INPUT_FILE1}]" in result.output
        assert f"[INFO] successfully generate document [{OUTPUT_MD_FILE2}] from template [{INPUT_FILE2}]" in result.output

    finally:
        shutil.copy(INPUT_MASTER_FILE, INPUT_FILE1)

